#ifndef MITOGEN_CONVOLUTION_H
#define MITOGEN_CONVOLUTION_H


//#include "image3d.h"
//#include "algo_export.h"
#include <complex>
#include <queue>
#include <fftw3.h>
#include <i3d/toolbox.h>

#ifdef GTGEN_WITH_MULTITHREADING
	#include <mutex>

	std::mutex i3dConvolution_External_FuncLock_st;
#endif

#include "../../macros.h"


#define is_odd(X) (((X) % 2) == 1)
#define is_even(X) (((X) % 2) == 0)


enum BoundaryConditionType
{
	BoundaryDirichlet,
	BoundaryNeumann
};


struct couple 
{
	  size_t value;
	  size_t position; 

	  couple(size_t v,size_t p):value(v),position(p) {};
	  couple() {};

	  bool operator<(const couple &x) const
	  {
			return ((this->value) < (x.value));
	  }
 };


template <class PRECISION>
size_t GetFFTWBufferSize(i3d::Vector3d<size_t> img_sz)
{
	 size_t buffer_line_length = 2 * (img_sz.x/2 + 1);

	 return (buffer_line_length * img_sz.y * img_sz.z * sizeof(PRECISION));
}


template <class T_in, class T_out, class PRECISION> double
estimate_needed_space(const i3d::Vector3d<size_t> &F_voi_sz,
							 const i3d::Vector3d<size_t> &G_voi_sz)
{
	 // If you want to perform convolution via FFT, you must align both
	 // tiles (their sizes are defined via VOIs) to the same size.
	 // This variable defines this new size that both tiles will 
	 // stretched to.
	 i3d::Vector3d<size_t> VOI_padded_dims = F_voi_sz + G_voi_sz - size_t(1);

	 // number of voxels in the padded image
	 double num_of_voxels =
				double(VOI_padded_dims.x) * // width
				double(VOI_padded_dims.y) * // height
				double(VOI_padded_dims.z); // depth

	 // Now, we will try to find out, how much memory is occupied by ONE
	 // stretched tile (size is defined in bytes).
	 double img_padded_sz_in = num_of_voxels *
				sizeof(T_in); // size of data type

	 double img_padded_sz_out = num_of_voxels *
				sizeof(T_out); // size of data type

	 // When performing Fourier transform (REAL->COMPLEX) we will also
	 // need a certain memory that will be occupied by the output, i.e.
	 // by the complex image.
	 double ft_img_padded_sz = num_of_voxels *
				sizeof(std::complex<PRECISION>)/2; // we store only half of the image

	 // There is some memory requirement by FFTW algorithm (an algorithm
	 // that performs Fast Fourier Transform - http://www.fftw.org/).
    double fftw_buffer_size = GetFFTWBufferSize<PRECISION>(VOI_padded_dims);

	 // The total size (excluding the input/output buffers) of the memory
	 // that will be needed for the evaluation of the convolution is the sum
	 // of all the previous variables. 
	 double total_size =
				2 * img_padded_sz_in +  // two convolved buffers
				1 * std::max(img_padded_sz_out, ft_img_padded_sz) +  
				1 * ft_img_padded_sz + // one Fourier domain
				1 * fftw_buffer_size; // one working buffer for FFTW

	 return total_size;
}


void generate_descending_sequence(const i3d::Vector3d<size_t> &dim,
										  std::queue<i3d::Vector3d<size_t>/**/> &seq)
{
	 // The input vector should be positive. It cannot contain zeros.
	 if (dim.x * dim.y * dim.z == 0)
		  throw ERROR_REPORT("Wrong dimensions");

	 // Pour the input vector into an auxiliary array
	 std::vector<couple> array(3);
	 array[0] = couple(dim.x, 0);
	 array[1] = couple(dim.y, 1);
	 array[2] = couple(dim.z, 2);

	 // Sort this array + remember the original positions
	 stable_sort(array.begin(), array.end());

	 // Store the initial value
	 seq.push(dim);

	 // Track the individual components (in descending order respecting
	 // their size) and decrease their value in successive steps.
	 for (int i=2; i>=0; i--)
	 {
		  do {
				array[i].value--;

				i3d::Vector3d<size_t> v;
				v[array[0].position] = array[0].value;
				v[array[1].position] = array[1].value;
				v[array[2].position] = array[2].value;

				// Store the new (decreased) triple
				seq.push(v);
		  } while (array[i].value > 1); // Decrease down to the value '1'
	 }
}


bool is_factorable(size_t number)
{
/*#ifdef I3D_DEBUG
	 size_t number_backup = number;
#endif*/
	 // possible primes: 2,3,5 and 7 -- for more, see documentation of FFTW
	 const size_t primes[] = {2,3,5,7,11,13};
	 const size_t sz = sizeof(primes)/sizeof(size_t);
	 size_t remainder;
	 size_t *powers = new size_t[sz];

	 for (size_t i=0; i<sz; i++)
	 {
		  powers[i] = 0;

		  do {
				remainder = number % primes[i];
				if (remainder == 0)
				{
					 number /= primes[i];
					 powers[i]++;
				}
		  } while (remainder == 0);
	 }

	 // there is no remainder and 
	 // the sum of powers of 11 and 13 gives '0' or '1'
	 if ((number == 1) && (powers[4] + powers[5] <= 1))
	 {
/*#ifdef I3D_DEBUG
		  cout << "Convolution: " << number_backup << " = ";
		  for (size_t j=0; j<sz; j++)
		  {
				if (j > 0)
					 cout << " * ";

				cout << primes[j] << "^" << powers[j];
		  }
		  cout << endl;
#endif*/
		  delete [] powers;
		  return true;
	 }

	 delete [] powers;
	 return false;
}


template < class T_in, class T_out, class PRECISION > bool
find_optimal_tile_size(const i3d::Vector3d<size_t> &img_dim,
							  const i3d::Vector3d<size_t> &ker_dim,
							  i3d::Vector3d<size_t> &tile_dim)
{
	 tile_dim = max(img_dim, ker_dim);

	 // How large portion of the total available memory can be dedicated
	 // to this process - we won't use the whole available memory, as we leave
	 // something for other processes etc...
	 const double memory_fraction = 0.3;
	 double free_ram = static_cast<double>(i3d::GetFreeMemory()); 
	 double mem_available = memory_fraction * free_ram;
	 double conv_size;

/*#ifdef I3D_DEBUG
	 cout << "Convolution: we can use: " << 
				human_readable(mem_available,0) << " bytes" << endl;
#endif*/

	 // NOTE:
	 // The memory for the input/output image and the convolution kernel has
	 // already been occupied, i.e. concerning I/O buffers we have
	 // nothing to do with. We will focus only on the memory requirements
	 // of FFTW and some local buffers.

 	 // Check if the convolution can be performed with the selected
	 // tiling of the input image and convolution kernel (the initial
	 // size tiles is the same as the whole image and the whole kernel
	 // i.e. without splitting). 
	 conv_size = estimate_needed_space<T_in, T_out, PRECISION>(img_dim,ker_dim);

/*#ifdef I3D_DEBUG
			  cout << "Convolution: we need: " << 
						 human_readable(conv_size,0) << " bytes" << endl;
#endif*/

	 // If tile fits the memory, leave the function and return 'true'.  
	 if (conv_size < mem_available)
	 {
/*#ifdef I3D_DEBUG
			  cout << "Convolution: enough memory for full convolution." << endl;
#endif*/
		  return true;
	 }

 	 /* If this code is reached, the available memory was found to be 
	  * too small. Prior to the convolution the filtered image has to 
	  * be split into tiles.
	  * 
	  * Some notes for better understanding:
	  * 
	  * - tile_dim ... The dimensions of the tile we're looking for
	  * 
	  * We will apply several limitations applied to `tile_dim`
	  *
	  * - all the tiles will be of the same size and shape 
	  *   (except for the boundary tiles)
	  * - both the image and the kernel will be split into the tiles
	  *   of the same size
	  * - padded `tile_dim` should be simply factored into small primes
	  * - `tile_dim` should be as large as possible (up to mem_available)
	  *
	  * The amount of the free memory is known. 
	  */

	 bool searching;
	 i3d::Vector3d<size_t> new_tile_dim;
	 std::queue<i3d::Vector3d<size_t>/**/> seq;
	 i3d::VOI<i3d::PIXELS> img_voi(0, img_dim), ker_voi(0, ker_dim);

	 generate_descending_sequence(tile_dim, seq);

	 do {
		  // Initially, we believe that the chosen size of the tile is OK,
		  // hence we do not need search for the next one.
		  searching = false; 
		 
		  // Try the largest possible tile.
		  new_tile_dim = seq.front();
		  seq.pop();

		  // Make an intersection of this tile this image and kernel to get
		  // the real size of tiles in image and kernel.
		  i3d::VOI<i3d::PIXELS> 
					 tile_voi(0, new_tile_dim),
					 img_tile_voi = tile_voi * img_voi,
					 ker_tile_voi = tile_voi * ker_voi;

		  // Compute the size of padded tile
		  i3d::Vector3d<size_t> new_tile_padded = 
					 img_tile_voi.size + ker_tile_voi.size - size_t(1);

		  // Check if the padded image can be factored into small primes
		  for (size_t idx=0; (idx<3) && (!searching); idx++)
		  {
				// Verify only the site (x,y or z), that is split
				if (new_tile_dim[idx] < tile_dim[idx])
				{
					 // If the padded tile cannot be factored, 
					 // try another (lower) tile
					 if (!is_factorable(new_tile_padded[idx]))
						  searching = true;
				}
		  } 

		  // Check if the tile size fits the available memory
		  conv_size = estimate_needed_space<T_in, T_out, PRECISION>(new_tile_dim,
																						new_tile_dim);

		  // If the memory requirements for the selected size of a tile 
		  // are more demanding than the available memory is, lower the size
		  // of tile, i.e. make the next loop
		  if (conv_size > mem_available)
				searching = true;
	 
	 } while (!seq.empty() && searching);

/*#if I3D_DEBUG
	 cout << "Convolution: original size: " << tile_dim << endl;
  	 cout << "Convolution: found optimal tile size: " << new_tile_dim << endl;
	 cout << "Convolution: memory: need x available: " << 
				human_readable(conv_size,0) << " x " << 
				human_readable(mem_available,0) << endl;
	 cout << "Convolution: size of data types: " << 
				2*sizeof(T_in) + 1*sizeof(T_out) + 3*sizeof(PRECISION) << endl;
#endif*/

	if (seq.empty())
	{
/*#ifdef I3D_DEBUG
		 cout << "Convolution: failed to find optimal tile size." << endl;
		 cout << "Convolution: ADVICE: extend your physical memory." << endl;
#endif*/

	 	 return false; // failed to allocate the memory 
	}

	tile_dim = new_tile_dim;
	return true;
}


template <class T>
void validate_images(const i3d::Image3d<T> &img_in,
							const i3d::Image3d<T> &kernel)
{
	 if ((img_in.GetImageSize() == 0) || (kernel.GetImageSize() == 0))
	 {
		  throw ERROR_REPORT("I3D::Convolution: empty images."); 
	 }

	 // both the input image and kernel should have the same resolution.
	 const i3d::Vector3d<float> acceptable_diff(0.001f, 0.001f, 0.001f);

	 i3d::Vector3d<float> 
				img_res = img_in.GetResolution().GetRes(),
				ker_res = kernel.GetResolution().GetRes();

	 i3d::Vector3d<float> real_diff(fabs(img_res.x - ker_res.x),
										fabs(img_res.y - ker_res.y),
										fabs(img_res.z - ker_res.z));

	 if (real_diff > acceptable_diff)
	 {
		  throw ERROR_REPORT("I3D::Convolution: images have "
									"different resolution."); 
	 }
}


void establish_location_of_tiles(const i3d::Vector3d<size_t> &image_dim,
											const i3d::Vector3d<size_t> &tile_dim,
											std::vector<i3d::VOI<i3d::PIXELS>/**/> &list_of_VOIs)
{
	 // How many pieces/tiles is split the general image into?
	 // (integer division with rounding up)
	 i3d::Vector3d<size_t> num_of_tiles;

	 num_of_tiles = (image_dim + tile_dim - size_t(1)) / tile_dim;

/*#ifdef I3D_DEBUG
	 cout << "Convolution: tiling: " << num_of_tiles << endl;
#endif */

	 // create the VOI that covers the whole image
	 i3d::VOI<i3d::PIXELS> voi_of_whole_img(0, image_dim);

	 // Create the lists of all the tiles into which the image will be split.
	 i3d::Vector3d<size_t> skip;

	 // Do it for the input image.
	 for (skip.x = 0; skip.x < num_of_tiles.x; skip.x++)
		  for (skip.y = 0; skip.y < num_of_tiles.y; skip.y++)
				for (skip.z = 0; skip.z < num_of_tiles.z; skip.z++)
				{
					 i3d::VOI<i3d::PIXELS> tile_voi(tile_dim * skip, tile_dim);
					 tile_voi *= voi_of_whole_img;

					 list_of_VOIs.push_back(tile_voi);
				}
}


template <class T>
void FFTshift(i3d::Image3d<T> &img, const i3d::Vector3d<int> &shift)
{
	 T *buffer_line;
	 size_t x,y,z;
	 size_t sz = img.GetSizeZ(),
			  sy = img.GetSizeY(),
			  sx = img.GetSizeX();

	 i3d::Vector3d<size_t> positive_shift;
	 positive_shift.x = size_t(shift.x % int(sx));
	 positive_shift.y = size_t(shift.y % int(sy));
	 positive_shift.z = size_t(shift.z % int(sz));

	 // x - axis
	 buffer_line = new T[sx];

	 for (z=0; z<sz; z++)
		  for (y=0; y<sy; y++)
		  {
				for (x=0; x<sx; x++)
					 buffer_line[x] = 
								*img.GetVoxelAddr((positive_shift.x + x) % sx, y, z);

				memcpy(img.GetVoxelAddr(0,y,z), buffer_line, sx*sizeof(T));
		  }

	 delete [] buffer_line;

	 // y-axis
	 buffer_line = new T[sy];

	 for (z=0; z<sz; z++)
		  for (x=0; x<sx; x++)
		  {
				for (y=0; y<sy; y++)
					 buffer_line[y] = 
								*img.GetVoxelAddr(x, (positive_shift.y + y) % sy, z);

				for (y=0; y<sy; y++)
					 *img.GetVoxelAddr(x,y,z) = buffer_line[y];
		  }
	 delete [] buffer_line;

	 // z-axis
	 buffer_line = new T[sz];

	 for (y=0; y<sy; y++)
		  for (x=0; x<sx; x++)
		  {
				for (z=0; z<sz; z++)
					 buffer_line[z] = 
								*img.GetVoxelAddr(x, y, (positive_shift.z + z) % sz);

				for (z=0; z<sz; z++)
					 *img.GetVoxelAddr(x,y,z) = buffer_line[z];
		  }
	 delete [] buffer_line;
}


template <class T> void* _fftw_malloc(size_t n) 
{
	 std::ostringstream os;
	 os << "i3d::fftw_malloc<" << 
				typeid(T).name() << ">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> void* _fftw_malloc<float>(size_t n)
{
	 return fftwf_malloc(n);
}
template <> void* _fftw_malloc<double>(size_t n)
{
	 return fftw_malloc(n);
}
template <> void * _fftw_malloc<long double>(size_t n)
{
	 return fftwl_malloc(n);
}


template <class T> struct _fftw_complex 
{ 
	 _fftw_complex()
	 {
		  std::ostringstream os;
		  os << "i3d::fftw_complex<" << typeid(T).name() << "> not implemented."; 
		  
		  throw ERROR_REPORT(os.str().c_str());
	 };
};
template <> struct _fftw_complex<float> 
{ 
	 typedef fftwf_complex type; 

	 type value;
};
template <> struct _fftw_complex<double> 
{ 
	 typedef fftw_complex type; 

	 type value;
};
template <> struct _fftw_complex<long double> 
{ 
	 typedef  fftwl_complex type; 

	 type value;
};


template <class T> struct _fftw_plan 
{
	 _fftw_plan()
	 {
		  std::ostringstream os;
		  os << "i3d::fftw_plan<" << typeid(T).name() << "> not implemented."; 
		  
		  throw ERROR_REPORT(os.str().c_str());
	 };
};
template <> struct _fftw_plan<float> 
{ 
	 typedef fftwf_plan type; 
	 type plan;

	 _fftw_plan(fftwf_plan input):plan(input) {};
};
template <> struct _fftw_plan<double> 
{ 
	 typedef fftw_plan type; 
	 type plan;

	 _fftw_plan(fftw_plan input):plan(input) {};
};
template <> struct _fftw_plan<long double> 
{ 
	 typedef fftwl_plan type; 
	 type plan;

	 _fftw_plan(fftwl_plan input):plan(input) {};
};


template <class T> _fftw_plan<T> _fftw_plan_dft_r2c_3d(int n0, int n1, int n2, T *in, _fftw_complex<T> *out, unsigned flags)
{
	 std::ostringstream os;
	 os << "i3d::fftw_plan_dft_r2c_3d<" << typeid(T).name() << 
				">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> _fftw_plan<float> _fftw_plan_dft_r2c_3d(int n0, int n1, int n2, float *in, _fftw_complex<float> *out, unsigned flags)
{
	 fftwf_plan plan = fftwf_plan_dft_r2c_3d(n0, n1, n2, in, 
											(fftwf_complex*) out,
											flags);

	 return _fftw_plan<float>(plan);
}
template <> _fftw_plan<double> _fftw_plan_dft_r2c_3d(int n0, int n1, int n2, double *in, _fftw_complex<double> *out, unsigned flags)
{
	 fftw_plan plan = fftw_plan_dft_r2c_3d(n0, n1, n2, in, 
										(fftw_complex *) out, 
										flags);

	 return _fftw_plan<double>(plan);
}
template <> _fftw_plan<long double> _fftw_plan_dft_r2c_3d(int n0, int n1, int n2, long double *in, _fftw_complex<long double> *out, unsigned flags)
{
	 fftwl_plan plan = fftwl_plan_dft_r2c_3d(n0, n1, n2, in, 
										(fftwl_complex *) out, 
										flags);

	 return _fftw_plan<long double>(plan);
}


template <class T> void _fftw_execute(const _fftw_plan<T> plan)
{
	 std::ostringstream os;
	 os << "i3d::fftw_execute<" << 
				typeid(T).name() << ">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());

}
template <> void _fftw_execute(const _fftw_plan<float> plan)
{
	 fftwf_execute(plan.plan);
}
template <> void _fftw_execute(const _fftw_plan<double> plan)
{
	 fftw_execute(plan.plan);
}
template <> void _fftw_execute(const _fftw_plan<long double> plan)
{
	 fftwl_execute(plan.plan);
}


template <class T> void _fftw_destroy_plan(_fftw_plan<T> plan)
{
	 std::ostringstream os;
	 os << "i3d::fftw_destroy_plan<" << 
				typeid(T).name() << ">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> void _fftw_destroy_plan(_fftw_plan<float> plan)
{
	 fftwf_destroy_plan(plan.plan);
}
template <> void _fftw_destroy_plan(_fftw_plan<double> plan)
{
	 fftw_destroy_plan(plan.plan);
}
template <> void _fftw_destroy_plan(_fftw_plan<long double> plan)
{
	 fftwl_destroy_plan(plan.plan);
}


template <class T> void _fftw_cleanup(void)
{
	 std::ostringstream os;
	 os << "i3d::fftw_cleanup<" << 
				typeid(T).name() << ">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> void _fftw_cleanup<float>(void)
{
	 fftwf_cleanup();
}
template <> void _fftw_cleanup<double>(void)
{
	 fftw_cleanup();
}
template <> void _fftw_cleanup<long double>(void)
{
	 fftwl_cleanup();
}


template <class T> void _fftw_free(void *memory)
{
	 std::ostringstream os;
	 os << "i3d::fftw_free<" << 
				typeid(T).name() << ">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> void _fftw_free<float>(void *memory)
{
	 fftwf_free(memory);
}
template <> void _fftw_free<double>(void *memory)
{
	 fftw_free(memory);
}
template <> void _fftw_free<long double>(void *memory)
{
	 fftwl_free(memory);
}


template <class S, class T> 
void RealImage2FFTWBuffer(const i3d::Image3d<S>& img, T* fftw_buffer, T factor)
{
	 i3d::Vector3d<size_t> img_sz = img.GetSize();
    // real -> multiplied by "2"!
	 size_t buffer_line_length = 2 * (img_sz.x/2 + 1);
	 size_t buffer_slice_sz = img_sz.y * buffer_line_length;

	 const S *it_img = img.GetFirstVoxelAddr();

	 for (size_t z=0; z<img_sz.z; z++)
		  for (size_t y=0; y<img_sz.y; y++)
		  {
				T *it_buffer = fftw_buffer 
						  + buffer_slice_sz*z 
						  + buffer_line_length*y;

				// track the line and copy the data
				for (size_t x=0; x<img_sz.x; x++)
				{
					 *it_buffer = static_cast<T>(*it_img) * factor;

					 it_buffer++;
					 it_img++;
				}
		  }
}

			
template < class T, class PRECISION>
void FFT(const i3d::Image3d<T> &real_img, 
			i3d::Image3d<std::complex<PRECISION>/**/> &complex_img,
			bool normalize = false)
{
//#ifdef WITH_FFTW

/*#ifdef I3D_DEBUG
	cout << "FFT: forward Fourier transform for real image: " << 
			  real_img.GetSize() << endl;
#endif*/

	
/*#ifdef GTGEN_WITH_MULTITHREADING
	// Note:
	// If the computer contains dual (or more) core CPU, we can utilize
	// multi-threading computing. The acceleration of computation is not 
	// enormous but not negligible. It was experimentally found out that
	// running two or more independent multi-threaded processes does not 
	// reduce the computation speed. Hence, each process can ask for 
	// running its task over all the available CPUs. On the other hand the
	// process should NOT ask for running N threads, where N > CPUs. It has
	// considerable (negative) impact on the performance.
	size_t CPUs = GetNumberOfProcessors();


	MUTEX_LOCK();

 	_fftw_init_threads<PRECISION>();
   _fftw_plan_with_nthreads<PRECISION>(CPUs);

	MUTEX_UNLOCK();
#endif*/

	 // the VOI covering the whole image area
	 i3d::Vector3d<size_t> sz = real_img.GetSize();

	 // allocate the new (complex) image of reduced size
	 complex_img.MakeRoom(sz.x/2 + 1, sz.y, sz.z);

#ifdef GTGEN_WITH_MULTITHREADING	
	// *** external function LOCK ***
	std::unique_lock<std::mutex> externalFuncLock(i3dConvolution_External_FuncLock_st);
#endif
	 
	 // allocate the required memory, if failed raise an exception
	 size_t buffer_size = GetFFTWBufferSize<PRECISION>(sz);
	 PRECISION *buffer = (PRECISION*) _fftw_malloc<PRECISION>(buffer_size);
	 
	 if (buffer == NULL)
	 {
		  std::ostringstream os;

		  os << "i3dlib::FourierTranform<" << typeid(PRECISION).name() <<
					 "> ... memory allocation failed.";
		  throw ERROR_REPORT(os.str().c_str());
	 }

	 // By default, the FT is not normalized. If required, each item is
	 // divided by factor (number of elements in the processed image).
	 PRECISION factor = 1;
	 if (normalize)
	 {
		  factor /= sqrt(static_cast<PRECISION>(real_img.GetImageSize()));
	 }
	 
	 // copy the input image data to the allocated (real) buffer and prepare
	 // the output (complex) buffer. Notice that these buffers point to the 
	 // same address. It means that the FFT will be computed 'in-place'.
	 RealImage2FFTWBuffer(real_img, buffer, factor); 
	 _fftw_complex<PRECISION> *cbuffer = (_fftw_complex<PRECISION> *) buffer; 
	
	 // prepare the FFTW plan for given data

	 
/*#ifdef GTGEN_WITH_MULTITHREADING
	 MUTEX_LOCK();
#endif*/

	 _fftw_plan<PRECISION> plan = 
		 _fftw_plan_dft_r2c_3d<PRECISION> (sz.z, sz.y, sz.x, 
													  buffer, cbuffer,
													  FFTW_ESTIMATE); 
													 

/*#ifdef GTGEN_WITH_MULTITHREADING
	 MUTEX_UNLOCK();
#endif*/

#ifdef GTGEN_WITH_MULTITHREADING	 
	 // *** external function unLOCK ***
	 externalFuncLock.unlock();
#endif

	// only execute function is thread safe (according to fftw documentation))
	
	 // perform the forward FFT
	 // NOTE: execute is only fftw function that is thread-safe
	 _fftw_execute<PRECISION>(plan);
	 
#ifdef GTGEN_WITH_MULTITHREADING	 
	 // *** external function LOCK ***
	 externalFuncLock.lock();
#endif

	 // release the memory allocated by FFTW plan
	 _fftw_destroy_plan<PRECISION> (plan); 
	 // NOTE: cleanup makes all existing plans undefined, that is undesirable
	 // when running concurrently
	 //_fftw_cleanup<PRECISION>();

/*#ifdef GTGEN_WITH_MULTITHREADING
	 _fftw_cleanup_threads<PRECISION>();
#endif*/
	 
	 // copy the results from FFTW-ready buffer into the image buffer
	 memcpy(complex_img.GetFirstVoxelAddr(),
			  cbuffer,
			  complex_img.GetImageSize() * sizeof(std::complex<PRECISION>));
			  
	 // release the memory occupied by the FFTW buffer.
	 _fftw_free<PRECISION>(buffer); 

/*#ifdef I3D_DEBUG
	cout << "FFT: forward Fourier transform finished" << endl;
#endif*/

/*#else // WITH_FFTW
  throw InternalException("i3dlib: FFT support is missing. "
								  "Enable WITH_FFTW flag and recompile the i3dlib.");
#endif // WITH_FFTW*/
}


template <class T> _fftw_plan<T> _fftw_plan_dft_c2r_3d(int n0, int n1, int n2, _fftw_complex<T> *in, T *out, unsigned flags)
{
	 std::ostringstream os;
	 os << "i3d::fftw_plan_dft_c2r_3d<" << typeid(T).name() << 
				">(...) not implemented.";

	 throw ERROR_REPORT(os.str().c_str());
}
template <> _fftw_plan<float> _fftw_plan_dft_c2r_3d(int n0, int n1, int n2, _fftw_complex<float> *in, float *out, unsigned flags)
{
	 fftwf_plan plan = fftwf_plan_dft_c2r_3d(n0, n1, n2, 
											(fftwf_complex *) in, out,
											flags);

	 return _fftw_plan<float>(plan);
}
template <> _fftw_plan<double> _fftw_plan_dft_c2r_3d(int n0, int n1, int n2, _fftw_complex<double> *in, double *out, unsigned flags)
{
	 fftw_plan plan = fftw_plan_dft_c2r_3d(n0, n1, n2, 
											(fftw_complex *) in, out,
											flags);

	 return _fftw_plan<double>(plan);
}
template <> _fftw_plan<long double> _fftw_plan_dft_c2r_3d(int n0, int n1, int n2, _fftw_complex<long double> *in, long double *out, unsigned flags)
{
	 fftwl_plan plan = fftwl_plan_dft_c2r_3d(n0, n1, n2, 
											(fftwl_complex *) in, out,
											flags);

	 return _fftw_plan<long double>(plan);
}


template <class S, class T> 
void FFTWBuffer2RealImage(const S* fftw_buffer, i3d::Image3d<T>& img, S factor)
{
	 i3d::Vector3d<size_t> img_sz = img.GetSize();
    // real - multiplied by "2"!
	 size_t buffer_line_length = 2 * (img_sz.x/2 + 1);
	 size_t slice_sz = img_sz.y * buffer_line_length;

	 T *it_img = img.GetFirstVoxelAddr();

	 if (std::numeric_limits<T>::is_integer)
	 {
		 for (size_t z=0; z<img_sz.z; z++)
			  for (size_t y=0; y<img_sz.y; y++)
			  {
					const S *it_buffer = fftw_buffer 
							  + slice_sz * z 
							  + buffer_line_length * y;
	
					// track the line and copy the data
					for (size_t x=0; x<img_sz.x; x++)
					{
						 *it_img = static_cast<T>(round(*it_buffer * factor));
	
						 it_buffer++;
						 it_img++;
					}
			  }
	 }
	 else
	 {
		 for (size_t z=0; z<img_sz.z; z++)
			  for (size_t y=0; y<img_sz.y; y++)
			  {
					const S *it_buffer = fftw_buffer 
							  + slice_sz * z 
							  + buffer_line_length * y;
	
					// track the line and copy the data
					for (size_t x=0; x<img_sz.x; x++)
					{
						 *it_img = static_cast<T>(*it_buffer * factor);
	
						 it_buffer++;
						 it_img++;
					}
			  }
	 }
}


template < class T, class PRECISION> 
void IFFT(const i3d::Image3d<std::complex<PRECISION>/**/> &complex_img, 
			 i3d::Image3d<T> &real_img,
			 bool normalize = false)
{
//#ifdef WITH_FFTW

/*#ifdef I3D_DEBUG
	cout << "IFFT: inverse Fourier transform for symmetric complex image " << 
			  complex_img.GetSize() << endl;
#endif*/

	
/*#ifdef ALGO_WITH_MULTITHREADING
	 // if required, use multi-threaded computing
	 size_t CPUs = GetNumberOfProcessors();

	 MUTEX_LOCK();

	 _fftw_init_threads<PRECISION>();
	 _fftw_plan_with_nthreads<PRECISION>(CPUs);

	 MUTEX_UNLOCK();
#endif*/


	 // the VOI covering the whole image area
	 i3d::Vector3d<size_t> c_sz = complex_img.GetSize();
	 i3d::Vector3d<size_t> r_sz = real_img.GetSize();

	 // verify that the sizes do match!
	 if ((c_sz.x != (r_sz.x/2 + 1)) || 
		  (c_sz.y != r_sz.y) ||
		  (c_sz.z != r_sz.z))
	 {
		  throw "IFFT error: the size of images does not match.";
	 }

#ifdef GTGEN_WITH_MULTITHREADING	
	// *** external function LOCK ***
	std::unique_lock<std::mutex> externalFuncLock(i3dConvolution_External_FuncLock_st);
#endif

	 // allocate the required memory, if failed raise an exception
	 _fftw_complex<PRECISION> *cbuffer = (_fftw_complex<PRECISION>*) 
				_fftw_malloc<PRECISION>(complex_img.GetImageSize() * 
												sizeof(std::complex<PRECISION>)); 
	 
	 if (cbuffer == NULL)
	 {
		  std::ostringstream os;

		  os << "i3dlib::FourierTranform<" << typeid(PRECISION).name() <<
					 "> ... memory allocation failed.";
		  throw ERROR_REPORT(os.str().c_str());
	 }
	 
	 // copy the input image data to the allocated (complex) buffer and prepare
	 // the output (real) buffer. Notice that these buffers point to the 
	 // same address. It means that the FFT will be computed 'in-place'.
	 memcpy(cbuffer,
			  complex_img.GetFirstVoxelAddr(),
			  complex_img.GetImageSize() * sizeof(std::complex<PRECISION>));

	 PRECISION *buffer = (PRECISION *) cbuffer; 
	
	 // prepare the FFT plan for given data
	 
/*#ifdef ALGO_WITH_MULTITHREADING
	 MUTEX_LOCK();
#endif*/

	 _fftw_plan<PRECISION> plan = 
		 _fftw_plan_dft_c2r_3d<PRECISION> (r_sz.z, r_sz.y, r_sz.x, 
													  cbuffer, buffer,
													  FFTW_ESTIMATE);
	 
/*#ifdef ALGO_WITH_MULTITHREADING
	 MUTEX_UNLOCK();
#endif*/

#ifdef GTGEN_WITH_MULTITHREADING	 
	 // *** external function unLOCK ***
	 externalFuncLock.unlock();
#endif

	  
	 // perform the inverse FT
	 // NOTE: execute is only fftw function that is thread-safe
	 _fftw_execute<PRECISION>(plan);

#ifdef GTGEN_WITH_MULTITHREADING	 
	 // *** external function LOCK ***
	 externalFuncLock.lock();
#endif
	 
	 // release the memory occupied by the FFT plan
	 _fftw_destroy_plan<PRECISION> (plan);	 
	 // NOTE: cleanup makes all existing plans undefined, that is undesirable
	 // when running concurrently
	 //_fftw_cleanup<PRECISION>();
	 
/*#ifdef ALGO_WITH_MULTITHREADING
	 _fftw_cleanup_threads<PRECISION>();
#endif*/
	 
	 // By default, the FT is not normalized. If required, each item is
	 // divided by factor (number of elements in the processed image).
	 PRECISION factor = 1;
	 if (normalize)
	 {
		  factor /= sqrt(static_cast<PRECISION>(real_img.GetImageSize()));
	 }
	
	 // copy the results from FFTW-like buffer to ordinary image
	 FFTWBuffer2RealImage(buffer, real_img, factor); 
	
	 // release the memory occupied by the FFTW buffer
	 _fftw_free<PRECISION>(cbuffer); 

/*#ifdef I3D_DEBUG
	cout << "IFFT: inverse Fourier transform finished" << endl;
#endif

#else // WITH_FFTW
  throw InternalException("i3dlib: FFT support is missing. "
								  "Enable WITH_FFTW flag and recompile the i3dlib.");
#endif // WITH_FFTW*/

}



template < class T_in, class T_out, class PRECISION > void
convolution_via_fft(const i3d::Image3d<T_in> &F,
						  const i3d::Image3d<T_in> &G,
						  i3d::Image3d<T_out> &H,
						  PRECISION factor)
{
	 // -------------- perform forward Fourier transforms ---------
	 // perform FFT over 'F'-part

	 i3d::Image3d<std::complex<PRECISION> > ft_F;
	 FFT(F, ft_F); 

/*#ifdef I3D_DEBUG
	 cout << "Convolution: entering critical section" << endl;
	 cout << "             you have: " << 
				human_readable(GetFreeMemory()) << " free bytes" << endl <<
				"             and you will use next: " <<
				human_readable(
					2*(GetFFTWBufferSize<PRECISION>(F.GetSize())),0) << 
				" bytes" << endl;
#endif*/

	 // perform FFT over 'G'-part
	 i3d::Image3d< std::complex<PRECISION> > ft_G;
	 FFT(G, ft_G); 
	 
	 // -------------- forward Fourier transforms completed --------

	 // Perform point-wise multiplication (including normalization)
	 std::complex<PRECISION> 
				*F_it = ft_F.GetFirstVoxelAddr(), 
				*G_it = ft_G.GetFirstVoxelAddr();

	 // The Fourier transform must be normalized as repetitive additions
	 // in FFT boost the signal. The solution is to divide the result
	 // by the number of image pixels.
	 // find out the size of padded tile (number of pixels)
	 factor /= F.GetImageSize(); 

	 // Perform the point-wise multiplication. Track the whole Fourier domain.
	 size_t half_sz = ft_F.GetImageSize();
	 for (size_t i=0; i<half_sz; i++)
	 {
		  *F_it *= (*G_it) * (factor);
		  F_it++;
		  G_it++;
	 }

	 ft_G.DisposeData();

	 // Perform inverse FFT - put the solution into the variable 'H'
	 H.MakeRoom(F.GetSize());
	 IFFT(ft_F, H);
}


template < class T_in, class T_out, class PRECISION > void
convolution_in_voi(const i3d::Image3d<T_in> &F,
						 const i3d::Image3d<T_in> &G,
						 i3d::Image3d<T_out> &H,
						 const i3d::VOI<i3d::PIXELS> &F_voi,
						 const i3d::VOI<i3d::PIXELS> &G_voi,
						 PRECISION factor,
						 BoundaryConditionType boundary_condition)
{

//#ifdef WITH_FFTW
	 /* F_voi_padding ... How much the 'F_voi' must be padded in order we
	  *                   could guarantee that the tiled-convolution is evaluated
	  *                   correctly, i.e. we won't lose any information.
	  * G_voi_padding ... Also G_voi must be padded. If 'G=G_voi' then padding
	  *                   of G is complementary to F, i.e. F padded and G padded
	  *                   have the same size.
	  * G_voi_halfsize ... One half of the area 'G_voi'
	  */
	 i3d::Vector3d<int> F_voi_padding = i3d::Vector3d<int>(G.GetSize() / size_t(2)),
						G_voi_padding = i3d::Vector3d<int>(F_voi.size / size_t(2)),
						G_voi_halfsize = i3d::Vector3d<int>(G_voi.size / size_t(2));

	 /* F_voi_padded is the base F_voi padded in all the directions
	  *
 	  *                                |-----------| 
	  * |-------|                      | |-------| |
	  * | F_voi | + F_voi_padding  =>  | | F_voi | | = F_voi_padded
	  * |-------|                      | |-------| |
	  *                                |-----------|
	  */
	 i3d::VOI<i3d::PIXELS> F_voi_padded(F_voi.offset - F_voi_padding,
									  F_voi.size + G.GetSize() - size_t(1));

	 /* G_voi_padded is the base G_voi padded ONLY in bottom-right direction.
	  * Theoretically, the padding should be done in the same way as in the
	  * case of F_voi. Fortunately, G_voi is cropped without any overlaps,
	  * i.e. everything beyond its boundaries is set to zero. For the further
	  * evaluation of FFT we need to shift G_voi to the origin. That is why
	  * we pad the G_voi in the following way. At least, we save some time.
	  *
	  *
 	  *                                |-------|---| 
	  * |-------|                      | G_voi |   |
	  * | G_voi | + G_voi_padding  =>  |-------|   |  = G_voi_padded
	  * |-------|                      |           |
	  *                                |-----------|
	  */
	 i3d::VOI<i3d::PIXELS> G_voi_padded(G_voi.offset/* - G_voi_padding*/, 
									  G_voi.size + F_voi.size - size_t(1));

	 /*
	  * If G=G_voi (is no G splitting) then F_active_area = F_voi_padded. 
	  * If G>G_voi, then we must select that part of F_voi_padded, that
	  * corresponds to the mirrored (across centre of F_voi_padded) and padded
	  * G_voi area. It is because convolution flips the kernel. As we handle
	  * with subparts of kernel G we must also manage this flipping.
	  *
	  * |---------------|                          |---------------|  
	  * |               |    G:|----|-------|      |               |
	  * |  F_voi_padded |      |    | G_voi |  =>  | F_voi_padded  |
	  * |               |      |    |-------|      |-------|       |
	  * |               |      |            |      |   *   |       |
	  * |---------------|      |------------|      |-------|-------|
	  *                                               ^ 
	  *                                               |
	  *                                            F_active_tile
	  */																 
	 i3d::VOI<i3d::PIXELS> F_active_tile(F_voi_padded.offset 
										+ i3d::Vector3d<int>(F_voi_padded.size)
										- i3d::Vector3d<int>(G_voi_padded.size)
										- G_voi.offset,
										G_voi_padded.size);

	 // NOTE:
	 // Some problems with memory fragmentation were found out. To avoid
	 // them we decided to allocate/deallocate the memory following the 
	 // idea First-In-Last-Out. This way, the memory block that will be released
	 // as the last one must be allocated as first. Similarly, the other
	 // blocks.

	 // ------------- prepare 'F'-part --------------
	 // pad the image
	 i3d::Image3d<T_in> F_tile_padded;
	 F_tile_padded.CopyFromVOI(F, F_active_tile, true);

	// If selected, apply the Neumann boundary condition to F
	if (boundary_condition == BoundaryNeumann)
	{
		for (int z = 0; z < int(F_tile_padded.GetSizeZ()); z++)
		{
			for (int y = 0; y < int(F_tile_padded.GetSizeY()); y++)
			{
				for (int x = 0; x < int(F_tile_padded.GetSizeX()); x++)
				{
					int abs_x = x + F_active_tile.offset.x;
					int abs_y = y + F_active_tile.offset.y;
					int abs_z = z + F_active_tile.offset.z;

					if (   (abs_x < 0) || (abs_x >= int(F.GetSizeX()))
					    || (abs_y < 0) || (abs_y >= int(F.GetSizeY()))
					    || (abs_z < 0) || (abs_z >= int(F.GetSizeZ())))
					{
						int bnd_x = std::min(std::max(abs_x, 0), int(F.GetSizeX()) - 1);
						int bnd_y = std::min(std::max(abs_y, 0), int(F.GetSizeY()) - 1);
						int bnd_z = std::min(std::max(abs_z, 0), int(F.GetSizeZ()) - 1);

						T_in copied_value = F.GetVoxel(bnd_x, bnd_y, bnd_z);
						F_tile_padded.SetVoxel(x, y, z, copied_value);
					}
				}
			}
		}
	}

	 // G_voi is a part of larger G. G originates in the top left corner
	 // while placing of G_voi starts from the opposite corner. As the
	 // dimensions of G_voi may be odd/even, we must shift the F data by
	 // one pixel in some cases.
	 i3d::Vector3d<int> rotation = G_voi_halfsize;

	 if (is_even(G_voi.size.x))
		  rotation.x--;

	 if (is_even(G_voi.size.y))
		  rotation.y--;

	 if (is_even(G_voi.size.z))
		  rotation.z--;

	 FFTshift(F_tile_padded, rotation);
	 // ------------- 'F'-part ready --------------
	

	 // ------------- prepare 'G'-part -------------
	 // pad the image
	 i3d::Image3d<T_in> G_tile_padded;

	 /*
	  * Initially, fill the image G_tile_padded with zeros (black background) 
	  * and then fill the top-left corner with values from the area in G defined 
	  * by G_voi.
	  **/
	 G_tile_padded.MakeRoom(G_voi_padded.size);
	 G_tile_padded.Clear(); 
	 
	 for (size_t z=0; z<G_voi.size.z; z++)
		  for (size_t y=0; y<G_voi.size.y; y++)
		  {
				const T_in* src = G.GetVoxelAddr(G_voi.offset.x,
															G_voi.offset.y + y,
															G_voi.offset.z + z);
				T_in* dst = G_tile_padded.GetVoxelAddr(0,y,z);
				memcpy(dst, src, G_voi.size.x * sizeof(T_in));
		  }
	 // ------------- 'G'-part ready -------------

	 // Here, the data is prepared in buffers 'F_tile_padded' & 'G_tile_padded'. 
	 // The both buffers are aligned to the same size i.e. 
	 // ready for convolution via Fourier transform.
	
	 // output buffer
	 i3d::Image3d<T_out> H_tile_padded;

	 convolution_via_fft(F_tile_padded,
								G_tile_padded,
								H_tile_padded,
								factor);

	 // Put the data to the output. Strictly speaking, fill the image H
	 // with the output data. Take care only of the area defined by 'F_voi'!
	 for (size_t z=0; z<F_voi.size.z; z++)
		  for (size_t y=0; y<F_voi.size.y; y++)
		  {
				T_out *H_padded_it = 
						  H_tile_padded.GetVoxelAddr(G_voi_halfsize.x,
															  G_voi_halfsize.y + y,
															  G_voi_halfsize.z + z),
						*H_it = H.GetVoxelAddr(F_voi.offset.x, 
													  F_voi.offset.y + y,
													  F_voi.offset.z + z);

				for (size_t i=0; i<F_voi.size.x; i++)
				{
					 *H_it += *H_padded_it;
					 H_it++; 
					 H_padded_it++;
				}
		  }

/*#else // WITH_FFTW
  throw InternalException("i3dlib: FFT support is missing. "
								  "Enable WITH_FFTW flag and recompile the i3dlib.");
#endif // WITH_FFTW*/
}


template < class PRECISION, class T_in, class T_out> void
Convolution_st (const i3d::Image3d<T_in> &img_in,
             const i3d::Image3d<T_in> &kernel,
             i3d::Image3d<T_out> &img_out,
             bool normalize_kernel = true,
             BoundaryConditionType boundary_condition = BoundaryDirichlet)
{
	if ((sizeof(T_out) > sizeof(PRECISION)) ||
		 (sizeof(T_in) > sizeof(PRECISION)))
	{
		 throw ERROR_REPORT("I3D::Convolution: Method precision is worse"
								  " than the precision of i/o images!?");
	}

/*#ifdef I3D_DEBUG
	 cout << "Convolution FFT based convolution started" << endl;
	 cout << "\t - input image size: " << img_in.GetSize() << endl;
	 cout << "\t - convolution kernel size: " << kernel.GetSize() << endl;
	 cout << "\t - required precision: " << typeid(PRECISION).name() << endl;
#endif*/

	 // Verify if the images (the image and the kernel) are suitable
	 // for convolution (non-empty, identical resolution)
	 validate_images(img_in, kernel);

	 // Allocate the memory for the output (including resolution)
	 // and clear the memory buffer (fill with zeros)
	 img_out.CopyMetaData(img_in);
	 img_out.Clear();

	 // If the output image is of integer datatype we cannot write the 
	 // results directly to its memory as we would loose the precision
	 // due to cumulative rounding error.
	 i3d::Image3d<PRECISION> img_store;
	 
	 if (std::numeric_limits<T_out>::is_integer)
	 {
		  img_store.CopyMetaData(img_in);
		  img_store.Clear();
	 }

	 // Find out the best memory allocation. There exist three conclusions:
	 // 1) memory is ok, i.e. the convolution can be performed directly
	 // 2) not enough memory which causes the convolved image 
	 //    splitting into tiles. In this case the tiles won't be of the same
	 //    size as the whole image or the whole kernel
	 // 3) not enough memory at all -> failure
	 i3d::Vector3d<size_t> tile_dim;

	 if (std::numeric_limits<T_out>::is_integer)
	 {
		  if (!find_optimal_tile_size<T_in, PRECISION, PRECISION>
				(img_in.GetSize(), kernel.GetSize(), tile_dim))
		  {
				throw ERROR_REPORT("I3D::Convolution: Not enough memory for"
										 " performing convolution.");
		  }
	 }
	 else
	 {
		  if (!find_optimal_tile_size<T_in, T_out, PRECISION>
				(img_in.GetSize(), kernel.GetSize(), tile_dim))
		  {
				throw ERROR_REPORT("I3D::Convolution: Not enough memory for"
										 " performing convolution.");
		  }
	 }

	 // Create the list of all possible tiles.
	 std::vector<i3d::VOI<i3d::PIXELS>/**/> tiles_from_image(0), tiles_from_kernel(0);

	 establish_location_of_tiles(img_in.GetSize(), tile_dim, tiles_from_image);
	 establish_location_of_tiles(kernel.GetSize(), tile_dim, tiles_from_kernel);

/*#ifdef I3D_DEBUG
	 cout << "Convolution: tiles of image:  " << 
				tiles_from_image.size() << endl;
	 cout << "Convolution: tiles of kernel: " << 
				tiles_from_kernel.size() << endl;
#endif*/

	 // The kernel normalization (sum == 1) might be required
	 PRECISION factor = 1;

	 if (normalize_kernel)
	 {
		  PRECISION sum = 0;
		  for (size_t i=0; i<kernel.GetImageSize(); i++)
				sum += kernel.GetVoxel(i);

		  factor /= sum; 

/*#ifdef I3D_DEBUG
		  cout << "Convolution: kernel normalization factor: " << factor << endl;
#endif*/

	 }

/*#ifdef I3D_DEBUG
	 cout << "Convolution: free memory (START): " << 
				human_readable(GetFreeMemory()) << endl;
#endif*/

	 // Perform the (tiled) convolution.
	 for (size_t i=0; i<tiles_from_image.size(); i++)
	 {
		  for (size_t j=0; j<tiles_from_kernel.size(); j++)
		  {
/*#ifdef I3D_DEBUG
		 	  cout << "Convolution: evaluating tile: " <<
					 i*tiles_from_kernel.size() + j + 1 << ". of " <<
					 tiles_from_image.size()*tiles_from_kernel.size() << 
					 " ## " << endl;
			  cout << "Convolution: " << 
					 "free memory when entering 'convolution_in_voi': " <<
					 human_readable(GetFreeMemory()) << endl;
#endif*/
			  // Perform the convolution. If the output image datatype
			  // is integer, the results cannot be forwarded the to the output
			  // image immediately. Otherwise, we would loose the precision.
			  // We must use the auxiliary image buffer "img_store" which is 
			  // of floating point datatype, i.e. it guarantees 
			  // more precise computation.
			  if (std::numeric_limits<T_out>::is_integer)
			  {
					convolution_in_voi(img_in, kernel, img_store,
											 tiles_from_image[i], tiles_from_kernel[j],
											 factor, boundary_condition);
			  }
			  else
			  {
					convolution_in_voi(img_in, kernel, img_out,
											 tiles_from_image[i], tiles_from_kernel[j],
											 factor, boundary_condition);
			  }

/*#ifdef I3D_DEBUG
		  	  cout << "Convolution: " << 
					 "free memory when leaving 'convolution_in_voi': " <<
					 human_readable(GetFreeMemory()) << endl;
#endif*/
	 
		  }
	 }
			  
	 if (std::numeric_limits<T_out>::is_integer)
	 {
		  // Copy the results stored in the floating point (precise) image
		  // into the integer image. During the copying the floating point
		  // numbers must be rounded and typecast.
		  T_out *it_out = img_out.GetFirstVoxelAddr();
		  PRECISION *it_store = img_store.GetFirstVoxelAddr();

		  for (size_t i=0; i<img_store.GetImageSize(); i++)
		  {
				*it_out = static_cast<T_out>(round(*it_store));
				it_out++;
				it_store++;
		  }
	 }

/*#ifdef I3D_DEBUG
	 cout << "Convolution: free memory (FINISH): " << 
				human_readable(GetFreeMemory()) << endl;
	 cout << "Convolution: FFT based convolution successfully finished." << endl;
#endif*/
}


//------------------------------------------------------------------------------
// explicit instantiations to force explicit build of the functions in this file
// because otherwise the linker couldn't find them
//------------------------------------------------------------------------------
// single precision
template  
void FFT(const i3d::Image3d<bool> &,
			i3d::Image3d<std::complex<float>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<i3d::GRAY8> &,
			i3d::Image3d<std::complex<float>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<i3d::GRAY16> &,
			i3d::Image3d<std::complex<float>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<float> &,
			i3d::Image3d<std::complex<float>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<double> &,
			i3d::Image3d<std::complex<float>/**/> &,
			bool);
			
// double precision
template  
void FFT(const i3d::Image3d<bool> &,
			i3d::Image3d<std::complex<double>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<i3d::GRAY8> &,
			i3d::Image3d<std::complex<double>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<i3d::GRAY16> &,
			i3d::Image3d<std::complex<double>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<float> &,
			i3d::Image3d<std::complex<double>/**/> &,
			bool);

template  
void FFT(const i3d::Image3d<double> &,
			i3d::Image3d<std::complex<double>/**/> &,
			bool);
			
			
// single precision			
template  
void IFFT(const i3d::Image3d<std::complex<float>/**/> &,
			 i3d::Image3d<i3d::GRAY8> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<float>/**/> &,
			 i3d::Image3d<i3d::GRAY16> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<float>/**/> &,
			 i3d::Image3d<float> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<float>/**/> &,
			 i3d::Image3d<double> &,
			 bool);
			 
// double precision
template  
void IFFT(const i3d::Image3d<std::complex<double>/**/> &,
			 i3d::Image3d<i3d::GRAY8> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<double>/**/> &,
			 i3d::Image3d<i3d::GRAY16> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<double>/**/> &,
			 i3d::Image3d<float> &,
			 bool);

template  
void IFFT(const i3d::Image3d<std::complex<double>/**/> &,
			 i3d::Image3d<double> &,
			 bool);
			
			
			
/*// fft shift
template  
void FFTshift(i3d::Image3d<bool> &img, const i3d::Vector3d<int> &shift);

template  
void FFTshift(i3d::Image3d<i3d::GRAY8> &img, const i3d::Vector3d<int> &shift);

template  
void FFTshift(i3d::Image3d<i3d::GRAY16> &img, const i3d::Vector3d<int> &shift);

template  
void FFTshift(i3d::Image3d<float> &img, const i3d::Vector3d<int> &shift);

template  
void FFTshift(i3d::Image3d<double> &img, const i3d::Vector3d<int> &shift);

// evaluate the amount of needed space
template  
size_t GetFFTWBufferSize<float>(i3d::Vector3d<size_t> img_sz);

template  
size_t GetFFTWBufferSize<double>(i3d::Vector3d<size_t> img_sz);*/




template
void Convolution_st<float, bool, float> (const i3d::Image3d <bool> &, 
									  const i3d::Image3d <bool> &, 
									  i3d::Image3d <float> &, bool, BoundaryConditionType);

template
void Convolution_st<double, bool, float> (const i3d::Image3d <bool> &, 
									   const i3d::Image3d <bool> &, 
									   i3d::Image3d <float> &, bool, BoundaryConditionType);

template
void Convolution_st<float, i3d::GRAY8, i3d::GRAY8> (const i3d::Image3d <i3d::GRAY8> &, 
									   const i3d::Image3d <i3d::GRAY8> &, 
									   i3d::Image3d <i3d::GRAY8> &, bool, BoundaryConditionType);

template  
void Convolution_st<float, i3d::GRAY16, i3d::GRAY16>(const i3d::Image3d <i3d::GRAY16> &, 
										const i3d::Image3d <i3d::GRAY16> &, 
										i3d::Image3d <i3d::GRAY16> &, bool, BoundaryConditionType);

template  
void Convolution_st<float, float, float> (const i3d::Image3d <float> &, 
									   const i3d::Image3d <float> &, 
									   i3d::Image3d <float> &, bool, BoundaryConditionType);

template  
void Convolution_st<double, i3d::GRAY8, i3d::GRAY8> (const i3d::Image3d <i3d::GRAY8> &, 
										const i3d::Image3d <i3d::GRAY8> &, 
										i3d::Image3d <i3d::GRAY8> &, bool, BoundaryConditionType);

template  
void Convolution_st<double, i3d::GRAY16, i3d::GRAY16>(const i3d::Image3d <i3d::GRAY16> &, 
										 const i3d::Image3d <i3d::GRAY16> &, 
										 i3d::Image3d <i3d::GRAY16> &, bool, BoundaryConditionType);

template  
void Convolution_st<double, float, float> (const i3d::Image3d <float> &, 
										const i3d::Image3d <float> &, 
										i3d::Image3d <float> &, bool, BoundaryConditionType);

template  
void Convolution_st<double, double, double> (const i3d::Image3d <double> &, 
										  const i3d::Image3d <double> &, 
										  i3d::Image3d <double> &, bool, BoundaryConditionType);


#endif
